/**************************************************************************************[MsSolver.cc]
  Copyright (c) 2022, Shiwei Pan
  **************************************************************************************************/

#include "MsSolver.h"
#include <vector>

#define __BMS__     15
#define __SC_NUM__  10
#define __SV_NUM__  50

const long min_step = 1000;
const long max_step = 100000000;
const int MY_RAND_MAX_INT = 10000000;
const float BASIC_SCALE = 0.0000001;

std::string _CutParenthesesNTail(std::string&& prettyFuncon) {
    auto pos = prettyFuncon.find('(');
    if(pos != std::string::npos) prettyFuncon.erase(prettyFuncon.begin()+pos, prettyFuncon.end());
	pos = prettyFuncon.find(' ');
	if(pos != std::string::npos) prettyFuncon.erase(prettyFuncon.begin(), prettyFuncon.begin() + pos + 1);
    return std::move(prettyFuncon);
}

void MsSolver::check_answer() {
    bool flag;
    for(unsigned int i = 0; i < ls_hard_cls.size(); i++) {
        flag = false;
        for(unsigned int j = 0; j < ls_hard_cls[i].size(); j++) {
            Lit& lit = ls_hard_cls[i][j];
            if((sign(lit) && !tmp_model[Minisat::var(lit)]) || (!sign(lit) && tmp_model[Minisat::var(lit)])) {
                flag = true;
                break;
            }
        }
        if(!flag) {
            reportf("check_answer failed\n"); 
            for(unsigned j = 0; j < ls_hard_cls[i].size(); j++) {
                Lit& lit = ls_hard_cls[i][j];
                reportf("%s%d %d\n", sign(lit) ? "-" : "", var(lit), tmp_model[var(lit)]);
            }
            exit(0);
        }
    }
    reportf("check_answer success\n");
    return;
}

void MsSolver::check_score() {
    memset(tmp_score.data(), 0, declared_n_vars * 2 * sizeof(long long));
    bool flag = false;
    for(int i = 0; i < declared_n_vars * 2; i++) {
        if(var_assump[i] || var_assump[i ^ 1]) continue;
        for(auto cls : lit_hard[i]) {
            if(hard_truth_num[cls] == 1 && ((i & 1) == !tmp_model[i / 2])) {
                tmp_score[i] += ls_hard_weight[cls];
            }
            else if(hard_truth_num[cls] == 0) {
                tmp_score[i] += ls_hard_weight[cls];
            }
        }
        for(auto cls : lit_soft[i]) {
            if(soft_truth_num[cls] == 1 && ((i & 1) == !tmp_model[i / 2])) {
                tmp_score[i] += ls_soft_weight[cls];
            }
            else if(soft_truth_num[cls] == 0) {
                tmp_score[i] += ls_soft_weight[cls];
            }
        }
        if(score[i] != tmp_score[i]) reportf("score wrong %d %lld %lld %d\n", i, score[i], tmp_score[i], tmp_model[i / 2]), flag = true;
    }
    if(flag) exit(0);
    return;
}

void MsSolver::settings() {
    rdprob = 0.01;
    rwprob = 0.1;
    smooth_probability = 0.01;
    if((total_soft_weight / ls_soft_cls.size()) > 10000) {
        h_inc = 300;
        softclause_weight_threshold = 500;
    }
    else {
        h_inc = 3;
        softclause_weight_threshold = 0;
    }
    if(declared_n_vars > 2000) {
        rdprob = 0.01;
        rwprob = 0.1;
        smooth_probability = 0.0000001;
    }
    hard_large_weight_num = 0;
    return;
}

void MsSolver::update_weight() {
    int cls;
    if (((rand() % MY_RAND_MAX_INT) * BASIC_SCALE) < smooth_probability && hard_large_weight_num > 0) {
        assert(soft_unsat.size() == soft_cls.size());
        for(int i = hard_unsat.get_strat(); i < hard_unsat.size(); i++) {
            cls = hard_unsat[i];
            if(ls_hard_weight[cls] > h_inc) {
                ls_hard_weight[cls] -= h_inc;
                if(hard_truth_num[cls] == 1) score[hard_sat_var[cls]] -= h_inc, var_score.update(Minisat::var(Minisat::toLit(hard_sat_var[cls])));
            }
            if(ls_hard_weight[cls] == 1) hard_large_weight_num--;
        }
        for(int i = soft_unsat.get_strat(); i < soft_unsat.size(); i++) {
            cls = soft_unsat[i];
            if(ls_soft_weight[cls] > 1) {
                ls_soft_weight[cls] -= 1;
                if(soft_truth_num[cls] == 1) score[soft_sat_var[cls]]--, var_score.update(Minisat::var(Minisat::toLit(soft_sat_var[cls])));
            }
        }
    }
    else {
        for(int i = 0; i < hard_unsat.get_strat(); i++) {
            cls = hard_unsat[i];
            if(ls_hard_weight[cls] == 1) hard_large_weight_num++;
            ls_hard_weight[cls] += h_inc;
            for(auto lit : ls_hard_cls[cls]) {
                if(var_assump[Minisat::toInt(lit)] || var_assump[Minisat::toInt(~lit)]) continue;
                score[Minisat::toInt(lit)] += h_inc;
                var_score.update(Minisat::var(lit));
            }
        }
        for(int i = 0; i < soft_unsat.get_strat(); i++) {
            cls = soft_unsat[i];
            if(ls_soft_weight[cls] < softclause_weight_threshold) {
                ls_soft_weight[cls]++;
                for(auto lit : ls_soft_cls[cls]) {
                    if(var_assump[Minisat::toInt(lit)] || var_assump[Minisat::toInt(~lit)]) continue;
                    score[Minisat::toInt(lit)]++;
                    var_score.update(Minisat::var(lit));
                }
            }
        }
    }
    return;
}

void MsSolver::flip(std::vector<int>& vars) {
    for(auto u : vars) {
        tmp_model[u] = !tmp_model[u];
        Lit tmp_lit = mkLit(u, tmp_model[u]);
        for(auto cls : lit_hard[toInt(tmp_lit)]) {
            if(hard_truth_num[cls] == -1 && hard_sat_var[cls] == -1) continue;
            hard_truth_num[cls]--;
            hard_sat_var[cls] -= toInt(tmp_lit);
            if(hard_truth_num[cls] == 0) {
                for(int i = 0; i < hard_length[cls]; i++) {
                    Lit lit = ls_hard_cls[cls][i];
                    if(lit == tmp_lit) continue;
                    score[toInt(lit)] += ls_hard_weight[cls];
                    var_score.update(Minisat::var(lit));
                }
            }
            else if(hard_truth_num[cls] == 1) {
                score[hard_sat_var[cls]] += ls_hard_weight[cls];
                var_score.update(Minisat::var(Minisat::toLit(hard_sat_var[cls])));
            }
            hard_unsat.update(cls);
        }
        for(auto cls : lit_soft[toInt(tmp_lit)]) {
            if(soft_truth_num[cls] == -1 && hard_sat_var[cls] == -1) continue;
            soft_truth_num[cls]--;
            soft_sat_var[cls] -= toInt(tmp_lit);
            if(soft_truth_num[cls] == 0) {
                for(int i = 0; i < soft_length[cls]; i++) {
                    Lit lit = ls_soft_cls[cls][i];
                    if(lit == tmp_lit) continue;
                    score[toInt(lit)] += ls_soft_weight[cls];
                    var_score.update(Minisat::var(lit));
                }
            }
            else if(soft_truth_num[cls] == 1) {
                score[soft_sat_var[cls]] += ls_soft_weight[cls];
                var_score.update(Minisat::var(Minisat::toLit(soft_sat_var[cls])));
            }
            soft_unsat.update(cls);
        }
        tmp_lit = mkLit(u, !tmp_model[u]);
        for(auto cls : lit_hard[toInt(tmp_lit)]) {
            if(hard_truth_num[cls] == -1 && hard_sat_var[cls] == -1) continue;
            if(hard_truth_num[cls] == 0) {
                for(int i = 0; i < hard_length[cls]; i++) {
                    Lit lit = ls_hard_cls[cls][i];
                    if(lit == tmp_lit) continue;
                    score[Minisat::toInt(lit)] -= ls_hard_weight[cls];
                    var_score.update(Minisat::var(lit));
                }
            }
            else if(hard_truth_num[cls] == 1) {
                score[hard_sat_var[cls]] -= ls_hard_weight[cls];
                var_score.update(Minisat::var(Minisat::toLit(hard_sat_var[cls])));
            }
            hard_truth_num[cls]++;
            hard_sat_var[cls] += toInt(tmp_lit);
            hard_unsat.update(cls);
        }
        for(auto cls : lit_soft[toInt(tmp_lit)]) {
            if(soft_truth_num[cls] == -1 && soft_sat_var[cls] == -1) continue;
            if(soft_truth_num[cls] == 0) {
                for(int i = 0; i < soft_length[cls]; i++) {
                    Lit lit = ls_soft_cls[cls][i];
                    if(lit == tmp_lit) continue;
                    score[Minisat::toInt(lit)] -= ls_soft_weight[cls];
                    var_score.update(Minisat::var(lit));
                }
            }
            else if(soft_truth_num[cls] == 1) {
                score[soft_sat_var[cls]] -= ls_soft_weight[cls];
                var_score.update(Minisat::var(Minisat::toLit(soft_sat_var[cls])));
            }
            soft_truth_num[cls]++;
            soft_sat_var[cls] += toInt(tmp_lit);
            soft_unsat.update(cls);
        }
        var_score.update(u);
    }
    return;
}

int MsSolver::select_by_BMS(int min_size, int id = -1) {
    long long max_score = 0;
    int v = -1;
    int tmp_v;
    long long tmp_score;
    int tmp_strat = var_score.get_strat();
    min_size = min(min_size, tmp_strat);
    for(int i = 0; i < min_size; i++) {
        tmp_v = var_score[rand() % tmp_strat];
        var_score.swap(tmp_v, --tmp_strat);
        if(tmp_v == id) continue;
        tmp_score = get_score(tmp_v, score, tmp_model);
        if(tmp_score > max_score) {
            max_score = tmp_score;
            v = tmp_v;
        }
        else if(tmp_score == max_score) {
            if(time_stamp[tmp_v] < time_stamp[v]) v = tmp_v;
        }
    }
    return v;
}

void MsSolver::pick_var(std::vector<int>& vars) {
    vars.clear();
    selected_var.clear();
    if(var_score.get_strat() > 0) {
        if ((rand() % MY_RAND_MAX_INT) * BASIC_SCALE < rdprob) vars.push_back(var_score[rand() % var_score.get_strat()]);
        else vars.push_back(select_by_BMS(__BMS__));
    }
    else {
        int tmp_strat, min_size, tmp_v;
        int max_v = -1;
        long long score_1 = INT_MAX;
        long long tmp_score, score_2, score_3;
        if((rand() % MY_RAND_MAX_INT) * BASIC_SCALE < rwprob){
            update_weight();
            int sel_c;
            if(hard_unsat.get_strat() > 0) {
                sel_c = hard_unsat[rand() % hard_unsat.get_strat()];
                vars.push_back(Minisat::var(ls_hard_cls[sel_c][rand() % hard_length[sel_c]]));
            }
            else {
                while(1){
                    sel_c = soft_unsat[rand() % soft_unsat.get_strat()];
                    if(soft_length[sel_c] > 0) break;
                }
                vars.push_back(Minisat::var(ls_soft_cls[sel_c][rand() % soft_length[sel_c]]));
            }
            return;
        }
        if(hard_unsat.get_strat() > 0) {
            min_size = min(__SC_NUM__, hard_unsat.get_strat());
            tmp_strat = hard_unsat.get_strat();
            for(int i = 0; i < min_size; i++) {
                int r = hard_unsat[rand() % tmp_strat];
                hard_unsat.swap(r, --tmp_strat);
                tmp_v = Minisat::var(ls_hard_cls[r][rand() % hard_length[r]]);
                if(selected[tmp_v]) continue;
                selected_var.push_back(tmp_v);
                selected[tmp_v] = true;
            }
        }
        else if(soft_unsat.get_strat() > 0) {
            min_size = min(__SC_NUM__, soft_unsat.get_strat());
            tmp_strat = soft_unsat.get_strat();
            for(int i = 0; i < min_size; i++) {
                int r = soft_unsat[rand() % tmp_strat];
                soft_unsat.swap(r, --tmp_strat);
                tmp_v = Minisat::var(ls_soft_cls[r][rand() % soft_length[r]]);
                if(selected[tmp_v]) continue;
                selected_var.push_back(tmp_v);
                selected[tmp_v] = true;
            }
        }
        for(auto u : selected_var) selected[u] = false;
        if(selected_var.size() == 1) {
            vars.push_back(selected_var[0]);
            return;
        }
        max_v = selected_var[0];
        score_1 = get_score(max_v, score, tmp_model);
        for(unsigned int i = 1; i < selected_var.size(); i++) {
            tmp_score = get_score(selected_var[i], score, tmp_model);
            if(tmp_score > score_1) {
                score_1 = tmp_score;
                max_v = tmp_v;
            }
            else if(tmp_score == score_1) {
                if(time_stamp[selected_var[i]] < time_stamp[max_v]) max_v = selected_var[i];
            }
        }
        score_2 = INT_MIN;
        std::vector<int> tmp_vec;
        int v1, v2;
        for(auto u : selected_var) {
            score_3 = get_score(u, score, tmp_model);
            tmp_vec.clear();
            tmp_vec.push_back(u); 
            flip(tmp_vec);
            //pseudo_flip(u);
            if(var_score.get_strat() > 1) {
                tmp_v = select_by_BMS(__SV_NUM__, u);
                tmp_score = get_score(tmp_v, score, tmp_model);
                //tmp_v = select_by_BMS2(__SV_NUM__, u);
                //tmp_score = get_score(tmp_v, score2, tmp_model);
                if(score_3 + tmp_score > 0) {
                    vars.push_back(u);
                    vars.push_back(tmp_v);
                    flip(tmp_vec);
                    return;
                }
                else if(score_3 + tmp_score > score_2) {
                    v1 = u;
                    v2 = tmp_v;
                    score_2 = score_3 + tmp_score;
                }
                else if(score_3 + tmp_score == score_2) {
                    if(time_stamp[u] + time_stamp[tmp_v] < time_stamp[v1] + time_stamp[v2]) v1 = u, v2 = tmp_v;
                }
            }
            flip(tmp_vec);
        }
        update_weight();
        if(score_1 > score_2) vars.push_back(max_v);
        else vars.push_back(v1), vars.push_back(v2);
    }
    return;
}

void MsSolver::local_search(vec<bool>& best_model, Int& goalvalue, Minisat::vec<Minisat::Lit>& assump_ps, Int UB_goalvalue, Int LB_goalvalue) {
    //reportf("local_search start\n");
    int no_improve_step = 0;
    Int ls_weight = 0;
    std::vector<int> vars;
    Minisat::vec<Minisat::Lit> soft_unsat_clause;
    if(ls_best_goalvalue > goalvalue) {
        best_model.copyTo(tmp_model);
        ls_step = 0;
        memset(time_stamp.data(), -1, time_stamp.size() * sizeof(int));
    }
    //check_answer();
    var_score.clear();
    hard_unsat.clear();
    soft_unsat.clear();
    for(unsigned int i = 0; i < ls_hard_cls.size(); i++) hard_truth_num[i] = 0, hard_sat_var[i] = 0;
    for(unsigned int i = 0; i < ls_soft_cls.size(); i++) soft_truth_num[i] = 0, soft_sat_var[i] = 0;
    memset(selected.data(), 0, selected.size() * sizeof(int));
    memset(score.data(), 0, score.size() * sizeof(long long));
    for(int i = 0; i < assump_ps.size(); i++) if(Minisat::var(assump_ps[i]) < declared_n_vars) var_assump[Minisat::toInt(assump_ps[i])] = 1;
    for(unsigned int i = 0; i < ls_hard_cls.size(); i++) {
        hard_length[i] = 0;
        for(unsigned int j = 0; j < ls_hard_cls[i].size(); j++) {
            Lit tmp_l = ls_hard_cls[i][j];
            if(var_assump[Minisat::toInt(tmp_l)]) {
                hard_sat_var[i] = -1;
                hard_truth_num[i] = -1;
                break;
            }
            else if(var_assump[Minisat::toInt(~tmp_l)]) continue;
            if((sign(tmp_l) && !tmp_model[Minisat::var(tmp_l)]) || (!sign(tmp_l) && tmp_model[Minisat::var(tmp_l)])) {
                hard_sat_var[i] += toInt(tmp_l);
                hard_truth_num[i]++;
            }
            std::swap(ls_hard_cls[i][j], ls_hard_cls[i][hard_length[i]++]);
        }
        if(hard_length[i] == 0) hard_sat_var[i] = -1, hard_truth_num[i] = -1;
    }
    for(unsigned int i = 0; i < ls_hard_cls.size(); i++) {
        if(hard_truth_num[i] == 1) score[hard_sat_var[i]] += ls_hard_weight[i];
        if(hard_truth_num[i] > -1) hard_unsat.insert(i);
    }
    for(unsigned int i = 0; i < ls_soft_cls.size(); i++) {
        soft_length[i] = 0;
        for(unsigned int j = 0; j < ls_soft_cls[i].size(); j++) {
            Lit tmp_l = ls_soft_cls[i][j];
            if(var_assump[Minisat::toInt(tmp_l)]) {
                soft_sat_var[i] = -1;
                soft_truth_num[i] = -1;
                break;
            }
            else if(var_assump[Minisat::toInt(~tmp_l)]) continue;
            std::swap(ls_soft_cls[i][j], ls_soft_cls[i][soft_length[i]++]);
            if((sign(tmp_l) && !tmp_model[Minisat::var(tmp_l)]) || (!sign(tmp_l) && tmp_model[Minisat::var(tmp_l)])) {
                soft_sat_var[i] += toInt(tmp_l);
                soft_truth_num[i]++;
            }
        }
        if(soft_length[i] == 0 && soft_sat_var[i] != -1) soft_sat_var[i] = -1, soft_truth_num[i] = -1, ls_weight += soft_cls[i].fst;
    }
    for(unsigned int i = 0; i < ls_soft_cls.size(); i++) {
        if(soft_truth_num[i] == 1) score[soft_sat_var[i]] += ls_soft_weight[i];
        else if(soft_truth_num[i] == 0) {
            for(unsigned int j = 0; j < ls_soft_cls[i].size(); j++) score[toInt(ls_soft_cls[i][j])] += ls_soft_weight[i];
        }
        if(soft_truth_num[i] > -1) soft_unsat.insert(i);
    }
    for(int i = 0; i < declared_n_vars; i++) if(!var_assump[Minisat::toInt(Minisat::mkLit(i, false))] && !var_assump[Minisat::toInt(Minisat::mkLit(i, true))]) var_score.insert(i);
    auto start_time = cpuTime();
    //while(no_improve_step < min_step && ls_step < max_step && cpuTime() < opt_cpu_lim - 20) {
    while(cpuTime() < opt_cpu_lim - 10 && cpuTime() - start_time < 300) {
        //printf("%d %s %s %d %d\n", ls_step, toString(currentvalue), toString(goalvalue), hard_unsat.get_strat(), var_score.get_strat());
        if(hard_unsat.get_strat() == 0) {
            //check_answer();
            currentvalue = fixed_goalval + ls_weight;
            for(int i = 0; i < soft_unsat.get_strat(); i++) currentvalue += soft_cls[soft_unsat[i]].fst;
            //currentvalue = evalGoal(soft_cls, tmp_model, soft_unsat_clause);
            //printf("sat %s\n", toString(currentvalue));
            if(currentvalue < best_goalvalue && currentvalue < goalvalue) {
                //reportf("\b currentvalue = %s %s %d %g\b\n", toString(currentvalue), toString(goalvalue), ls_step, cpuTime());
                local_update++;
                goalvalue = currentvalue;
                tmp_model.copyTo(best_model);
                ls_best_goalvalue = currentvalue;
                if (goalvalue < UB_goalvalue && opt_output_top < 0) UB_goalvalue = goalvalue;
                if (UB_goalvalue == LB_goalvalue) return;
            }
            if(soft_unsat.get_strat() == 0) return;
        }
        pick_var(vars);
        for(unsigned int i = 0; i < vars.size(); i++) time_stamp[vars[i]] = ls_step;
        flip(vars);
        no_improve_step++;
        ls_step++;
    }
    for(int i = 0; i < assump_ps.size(); i++) if(Minisat::var(assump_ps[i]) < declared_n_vars) var_assump[Minisat::toInt(assump_ps[i])] = 0;
    for(unsigned int i = 0; i < origin_assigns.size(); i++) var_assump[origin_assigns[i]] = 1;
    return;        
}
